# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""RIOT Driver for RIOT PAL
This module handles generic connection and IO to the make term.
"""
import logging
import os
import pexpect


class RiotDriver:
    """Contains all reusable functions for connecting, sending and receiving
    data.
    """

    def __init__(self, timeout=5, path=None):
        if path is not None:
            os.chdir(path)
        self.child = pexpect.spawnu("make term", timeout=timeout,
                                    codec_errors='replace')

    def close(self):
        """Closes the spawned process"""
        self.child.close()

    def readline(self):
        """Reads a line from a make term process and strips away all additional
        data so only the output of the device is left."""
        try:
            response = self.child.readline()
            response = response.split('# ', 1)[-1]
            response = response.replace('\n', '')
            response = response.replace('\r', '')
        except (ValueError, TypeError, pexpect.TIMEOUT) as exc:
            logging.debug(exc)
            raise TimeoutError
        logging.debug("Response: %s", response)
        return response

    def write(self, data):
        """Tries write data and adds a newline."""
        logging.debug("Writing: %s", data)
        self.child.write(data + '\n')



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""Device Under Tests Shell for RIOT PAL
This module handles parsing of information from RIOT shell base tests.
"""
import logging
import json
try:
    from .base_device import BaseDevice
except ImportError:
    from base_device import BaseDevice

RESULT_SUCCESS = 'Success'
RESULT_ERROR = 'Error'
RESULT_TIMEOUT = 'Timeout'


class ShellParser:
    """Parses commands and resposes from the shell."""
    COMMAND = 'Command: '
    SUCCESS = 'Success: '
    ERROR = 'Error: '
    TIMEOUT = 'Timeout: '

    def __init__(self, dev):
        self.dev = dev

    @staticmethod
    def _try_parse_data(data):
        if ('[' in data) and (']' in data):
            parsed_data = []
            data = data[data.find("[")+1:data.find("]")]
            data_list = data.split(', ')
            for value in data_list:
                try:
                    parsed_data.append(int(value, 0))
                except ValueError:
                    parsed_data.append(value)
            logging.debug(parsed_data)
            return parsed_data
        return None

    def send_and_parse_cmd(self, send_cmd):
        """Returns a dictionary with information from the event.

        Returns:
            dict:
            The return hold dict values in the following keys::
                msg - The message from the response, only used for information.
                cmd - The command sent, used to track what has occured.
                data - Parsed information of the data requested.
                result - Either success, error or timeout.
        """
        # pylint: disable=W0212
        self.dev._write(send_cmd)
        # pylint: disable=W0212
        try:
            cmd_info = {'cmd': send_cmd, 'data': None}
            response = self.dev._readline()
            while response != '':
                if self.COMMAND in response:
                    cmd_info['msg'] = response.replace(self.COMMAND, '')
                    cmd_info['cmd'] = cmd_info['msg'].replace('\n', '')

                if self.SUCCESS in response:
                    clean_msg = response.replace(self.SUCCESS, '')
                    cmd_info['msg'] = clean_msg.replace('\n', '')
                    cmd_info['result'] = RESULT_SUCCESS
                    cmd_info['data'] = self._try_parse_data(cmd_info['msg'])
                    break

                if self.ERROR in response:
                    clean_msg = response.replace(self.ERROR, '')
                    cmd_info['msg'] = clean_msg.replace('\n', '')
                    cmd_info['result'] = RESULT_ERROR
                    break
                # pylint: disable=W0212
                response = self.dev._readline()
        except TimeoutError:
            cmd_info['result'] = RESULT_TIMEOUT
            logging.debug(RESULT_TIMEOUT)

        return cmd_info


class JSONParser:
    """Handles parsing of specific json data

    Args:
        dev -> device to connect send and recieve data
    """

    def __init__(self, dev):
        self.dev = dev

    def _send_cmd(self, cmd_to_send, end_key='result'):
        # pylint: disable=W0212
        self.dev._write(cmd_to_send)
        cmd_info = {}
        while end_key not in cmd_info:
            line = ""
            try:
                # pylint: disable=W0212
                line = self.dev._readline()
                cmd_info.update(json.loads(line))
            except json.decoder.JSONDecodeError:
                if 'msg' not in cmd_info:
                    cmd_info['msg'] = []
                cmd_info['msg'].append(line)
            except TimeoutError:
                cmd_info['result'] = RESULT_TIMEOUT
        return cmd_info

    def send_and_parse_cmd(self, cmd_to_send):
        """Returns a dictionary with information from the event
        Args:
            send_cmd(str): The command to write to the device
            to_byte_array: If True and data is bytes leave it as an array
            timeout: Optional timeout value for command specific timeouts
        Returns:
            dict:
            The return hold dict values in the following keys::
            msg - The message from the response, only used for information.
            cmd - The command sent, used to track what has occured.
            data - Parsed information of the data requested.
            result - Either success, error or timeout.
        """

        cmd_info = {'cmd': cmd_to_send}
        cmd_info.update(self._send_cmd(cmd_to_send))
        return cmd_info


class DutShell:
    """Device Under Test shell class
    Args:
        parser(str): Selects the parser to use {shell, json}
    """

    def __init__(self, *args, **kwargs):
        self.parser = None

        parser = kwargs.pop('parser', 'shell')

        self.dev = BaseDevice(*args, **kwargs)
        if parser == 'shell':
            self.parser = ShellParser(self.dev)
        elif parser == 'json':
            # pylint: disable=R0204
            self.parser = JSONParser(self.dev)
        else:
            raise NotImplementedError()

    def send_cmd(self, cmd_to_send, *args, **kwargs):
        return self.parser.send_and_parse_cmd(cmd_to_send, *args, **kwargs)



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""Abstract Base Device for RIOT PAL
This module is the abstract device that interfaces to a driver.  It handles the
selection and initialzation of the driver.

Example:
    class ApplicationDevice(BaseDevice):
        def show_example(self):
            self._write('Send Example')
            print(self._read())

    example_use_case = ApplicationDevice('serial', port='my/port/name')
    example_use_case.show_example()
"""
import logging
try:
    from .serial_driver import SerialDriver
except ImportError:
    from serial_driver import SerialDriver
try:
    from .riot_driver import RiotDriver
except ImportError:
    from riot_driver import RiotDriver


class BaseDevice:
    """Instance for devices to connect and utilize drivers.

    Args:
        driver_type(str): Selects the driver that is used on the devices.
            'serial' uses the standard serial port, all following arguments
            get passed through.
            'riot' uses the riot make term system.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """

    def __init__(self, *args, **kwargs):
        self._driver = self._driver_from_config(*args, **kwargs)

    def close(self):
        """Closes the device connection."""
        self._driver.close()

    def _readline(self):
        """Reads data from the driver.

        Returns:
            str: string of data if success, driver defined error if failed.
        """
        return self._driver.readline()

    def _write(self, data):
        """Writes data to the driver.

        Args:
            data(str): Variable length argument list.
        """
        return self._driver.write(data)

    @staticmethod
    def _driver_from_config(*args, **kwargs):
        """Returns driver instance given configuration"""
        driver_type = kwargs.pop('driver_type', 'serial')

        if driver_type == 'serial':
            return SerialDriver(*args, **kwargs)
        elif driver_type == 'riot':
            return RiotDriver(*args, **kwargs)
        elif driver_type == 'driver':
            return kwargs['driver']
        raise NotImplementedError()

    @classmethod
    def copy_driver(cls, device):
        """Copies the driver instance so many devices can use one driver."""
        # pylint: disable=W0212
        logging.debug("Cloning Driver: %r", device._driver)
        return cls(driver_type='driver', driver=device._driver)



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""packet init for RIOT PAL
This exposes usful modules in the RIOT PAL packet
"""
from .dut_shell import DutShell

__all__ = ['DutShell']



#! /usr/bin/env python3
# Copyright (c) 2019 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT

"""
This script handles interfacing to riot DUT firmware.  It parses data and
exposes from the dut_shell class to be run in a shell.

The purpose of this script is allow easy setup and manual usage of the DUT.

Usage
-----

```
usage: dut_pyshell.py   [-h]
                        [--loglevel {debug,info,warning,error,fatal,critical}]
                        [--port PORT]
                        [--raw-data]

optional arguments:
  --help, -h
                        show this help message and exit
  --loglevel, -l
                        {debug,info,warning,error,fatal,critical}
                        Python logger log level (default: warning)
  --port, -p
                        Specify the serial port
  --rawdata, -r
                        Shows unfilted data, usually raw json
                        (default: False)
```
"""
import cmd
from json import dumps
import logging
import argparse

try:
    import readline
except ImportError:
    readline = None
import serial.tools.list_ports
try:
    from .dut_shell import DutShell, RESULT_SUCCESS
except ImportError:
    from dut_shell import DutShell, RESULT_SUCCESS


class DutPyShell(cmd.Cmd):
    """Command loop for the PHiLIP interface

    Args:
        port - Serial port for the PHiLIP, if None connection wizard tries to
               connent
        data_only - If true only data prints from command an not the whole
                    response struct
    """
    prompt = 'node: '

    def __init__(self, port=None, rawdata=False):
        if port is None:
            self.dut = self._connect_wizard()
        else:
            self.dut = DutShell(port, parser='json')
        self.cmd_list = self.dut.send_cmd('help')['data']
        self.data_only = not rawdata
        cmd.Cmd.__init__(self)

    @staticmethod
    def _connect_wizard():
        print("Starting DUT python shell")
        serial_devices = serial.tools.list_ports.comports()
        if len(serial_devices) == 0:
            raise ConnectionError("Could not find any available devices")
        elif len(serial_devices) == 1:
            print('Connected to {}'.format(serial_devices[0]))
            return DutShell(port=serial_devices[0][0], parser='json')
        else:
            print('Select a serial port:')
            for i, s_dev in enumerate(serial_devices):
                print("{}: {}".format(i, s_dev))
            s_num = int(input("Selection(number): "))
            return DutShell(port=serial_devices[int(s_num)][0], parser='json')

    def preloop(self):
        """Used to get the history of commands"""
        if readline:
            try:
                readline.read_history_file()
            except IOError:
                pass

    def do_send_cmd(self, arg):
        """Sends a command to the shell

        Usage:
            send_cmd <cmd_name> [args]

        Args:
            cmd_name: The name of the command
            args: Arguements for the command

        """
        try:
            results = self.dut.send_cmd(arg)
        except KeyError as exc:
            print('Could not parse argument {}'.format(exc))
        except (TypeError, ValueError, SyntaxError) as exc:
            print(exc)
        else:
            self._print_func_result_success(results)

    def complete_send_cmd(self, text, line, begidx, endidx):
        """Completes arg with command list"""
        begidx = begidx
        endidx = endidx
        return self._complete_cmd_list(text, line)

    def _complete_cmd_list(self, text, line):
        mline = line.partition(' ')[2]
        offs = len(mline) - len(text)
        return [s[offs:] for s in self.cmd_list if s.startswith(mline)]

    def do_rawdata(self, arg):
        """Sets raw or filtered data

        Usage:
            rawdata [val]

        Args:
            val: if 'on' then raw data, if 'off' than filtered
        """
        if arg:
            if arg.upper() == "OFF":
                self.data_only = True
                print("Filtering for data")
            elif arg.upper() == "ON":
                self.data_only = False
                print("Raw data, no filtering")
            else:
                print("Incorrect arg")
        elif self.data_only:
            self.data_only = False
            print("Raw data, no filtering")
        else:
            self.data_only = True
            print("Filtering for data")

    @staticmethod
    def complete_rawdata(text, line, begidx, endidx):
        """Completes arg filter option"""
        begidx = begidx
        endidx = endidx
        mline = line.partition(' ')[2]
        offs = len(mline) - len(text)
        filter_options = ['on', 'off']
        return [s[offs:] for s in filter_options if s.startswith(mline)]

    @staticmethod
    def do_exit(arg):
        """I mean it should be obvious

        Usage:
            exit
        """
        arg = arg
        return True

    def _print_func_result_success(self, results):
        if not isinstance(results, list):
            results = [results]
        result = RESULT_SUCCESS
        printed_something = False
        for res in results:
            if self.data_only and 'result' in res:

                if res['result'] == RESULT_SUCCESS:
                    if 'data' in res:
                        for data in res['data']:
                            print(dumps(data))
                            printed_something = True
                else:
                    result = res['result']
            else:
                print(dumps(res))
                printed_something = True
        if not printed_something:
            print(result)

    def _print_func_result(self, func, arg):
        values = (arg or '').split(' ')
        func_args = [v for v in values if v]
        try:
            results = func(*func_args)
        except KeyError as exc:
            print('Could not parse argument {}'.format(exc))
        except (TypeError, ValueError, SyntaxError) as exc:
            print(exc)
        else:
            self._print_func_result_success(results)


def _exit_cmd_loop():
    if readline:
        try:
            readline.write_history_file()
        except IOError:
            pass


def main():
    """Main program"""

    parser = argparse.ArgumentParser()

    log_levels = ('debug', 'info', 'warning', 'error', 'fatal', 'critical')
    parser.add_argument('--loglevel', '-l', choices=log_levels,
                        default='warning', help='Python logger log level')
    parser.add_argument('--port', '-p', help='Specifies the serial port',
                        default=None)
    parser.add_argument('--rawdata', '-r', default=False,
                        action='store_true',
                        help='Shows unfilted data, usually raw json')
    pargs = parser.parse_args()

    logging.basicConfig(level=getattr(logging, pargs.loglevel.upper()))
    try:
        DutPyShell(port=pargs.port, rawdata=pargs.rawdata).cmdloop()
        _exit_cmd_loop()
    except KeyboardInterrupt:
        _exit_cmd_loop()


if __name__ == '__main__':
    main()
